@@ -2,7 +2,7 @@ from pathlib import Path
 
 import click
 
-from ggshield.core.utils import handle_exception
+from ggshield.core.utils import ScanContext, ScanMode, handle_exception
 from ggshield.output import OutputHandler
 from ggshield.scan.docker import docker_scan_archive
 
@@ -24,14 +24,18 @@ def docker_archive_cmd(
     config = ctx.obj["config"]
     output_handler: OutputHandler = ctx.obj["output_handler"]
 
+    scan_context = ScanContext(
+        scan_mode=ScanMode.DOCKER,
+        command_path=ctx.command_path,
+    )
+
     try:
         scan = docker_scan_archive(
             archive=archive,
             client=ctx.obj["client"],
             cache=ctx.obj["cache"],
-            verbose=config.verbose,
             matches_ignore=config.secret.ignored_matches,
-            scan_id=str(archive),
+            scan_context=scan_context,
             ignored_detectors=config.secret.ignored_detectors,
         )
 
