@@ -1,7 +1,7 @@
 import click
 
 from ggshield.core.git_shell import get_list_commit_SHA
-from ggshield.core.utils import SupportedScanMode, handle_exception
+from ggshield.core.utils import ScanContext, ScanMode, handle_exception
 from ggshield.scan.repo import scan_commit_range
 
 
@@ -23,16 +23,19 @@ def range_cmd(ctx: click.Context, commit_range: str) -> int:  # pragma: no cover
         if config.verbose:
             click.echo(f"Commits to scan: {len(commit_list)}", err=True)
 
+        scan_context = ScanContext(
+            scan_mode=ScanMode.COMMIT_RANGE,
+            command_path=ctx.command_path,
+        )
+
         return scan_commit_range(
             client=ctx.obj["client"],
             cache=ctx.obj["cache"],
             commit_list=commit_list,
             output_handler=ctx.obj["output_handler"],
-            verbose=config.verbose,
             exclusion_regexes=ctx.obj["exclusion_regexes"],
             matches_ignore=config.secret.ignored_matches,
-            scan_id=commit_range,
-            mode_header=SupportedScanMode.COMMIT_RANGE.value,
+            scan_context=scan_context,
             ignored_detectors=config.secret.ignored_detectors,
         )
     except Exception as error:
