@@ -1,30 +1,32 @@
+import codecs
 import concurrent.futures
 import logging
 import re
+from ast import literal_eval
+from concurrent.futures import Future
 from dataclasses import dataclass
 from pathlib import Path
-from typing import Any, Callable, Dict, Iterable, List, NamedTuple, Optional, Set, Tuple
+from typing import Callable, Dict, Iterable, List, NamedTuple, Optional, Set, Tuple
 
+import charset_normalizer
 import click
 from pygitguardian import GGClient
-from pygitguardian.config import MULTI_DOCUMENT_LIMIT
+from pygitguardian.config import DOCUMENT_SIZE_THRESHOLD_BYTES, MULTI_DOCUMENT_LIMIT
 from pygitguardian.models import Detail, ScanResult
 
 from ggshield.core.cache import Cache
-from ggshield.core.constants import CPU_COUNT, MAX_FILE_SIZE
-from ggshield.core.extra_headers import generate_command_id, get_extra_headers
+from ggshield.core.extra_headers import get_headers
 from ggshield.core.filter import (
     is_filepath_excluded,
     remove_ignored_from_result,
     remove_results_from_ignore_detectors,
 )
 from ggshield.core.git_shell import GIT_PATH, shell
-from ggshield.core.text_utils import STYLE, format_text
+from ggshield.core.text_utils import STYLE, display_error, format_text, pluralize
 from ggshield.core.types import IgnoredMatch
-from ggshield.core.utils import REGEX_HEADER_INFO, Filemode
+from ggshield.core.utils import REGEX_HEADER_INFO, Filemode, ScanContext
 
 from ..iac.models import IaCScanResult
-from .scannable_errors import handle_scan_chunk_error
 
 
 logger = logging.getLogger(__name__)
@@ -178,29 +180,54 @@ class ScanCollection(NamedTuple):
 class File:
     """Class representing a simple file."""
 
-    def __init__(self, document: str, filename: str):
-        self.document = document
+    def __init__(self, document: Optional[str], filename: str):
+        self._document = document
         self.filename = filename
         self.filemode = Filemode.FILE
 
     def relative_to(self, root_path: Path) -> "File":
-        return File(self.document, str(Path(self.filename).relative_to(root_path)))
+        return File(self._document, str(Path(self.filename).relative_to(root_path)))
+
+    @property
+    def document(self) -> str:
+        self.prepare()
+        assert self._document is not None
+        return self._document
+
+    def prepare(self) -> None:
+        """Ensures self._document has been decoded"""
+        if self._document is not None:
+            return
+        with open(self.filename, "rb") as f:
+            self._document = File._decode_bytes(f.read(), self.filename)
 
     @staticmethod
     def from_bytes(raw_document: bytes, filename: str) -> "File":
-        document = raw_document.decode(errors="replace")
+        """Creates a File instance for a raw document. Document is decoded immediately."""
+        document = File._decode_bytes(raw_document, filename)
         return File(document, filename)
 
-    @property
-    def scan_dict(self) -> Dict[str, Any]:
-        """Return a payload compatible with the scanning API."""
-        return {
-            "filename": self.filename
-            if len(self.filename) <= 256
-            else self.filename[-255:],
-            "document": self.document,
-            "filemode": self.filemode,
-        }
+    @staticmethod
+    def from_path(filename: str) -> "File":
+        """Creates a File instance for a file. Content is *not* read immediately."""
+        return File(None, filename)
+
+    @staticmethod
+    def _decode_bytes(raw_document: bytes, filename: str) -> str:
+        """Low level function to decode bytes, tries hard to find the correct encoding.
+        For now it returns an empty string if the document could not be decoded"""
+        result = charset_normalizer.from_bytes(raw_document).best()
+        if result is None:
+            # This means we were not able to detect the encoding. Report it using logging for now
+            # TODO: we should report this in the output
+            logger.warning("Skipping %s, can't detect encoding", filename)
+            return ""
+
+        # Special case for utf_8 + BOM: `bytes.decode()` does not skip the BOM, so do it
+        # ourselves
+        if result.encoding == "utf_8" and raw_document.startswith(codecs.BOM_UTF8):
+            raw_document = raw_document[len(codecs.BOM_UTF8) :]
+        return raw_document.decode(result.encoding, errors="replace")
 
     def __repr__(self) -> str:
         return f"<File filename={self.filename} filemode={self.filemode}>"
@@ -238,10 +265,6 @@ class Files:
         """Convenience property to list filenames in the same order as files"""
         return [x.filename for x in self.files]
 
-    @property
-    def scannable_list(self) -> List[Dict[str, Any]]:
-        return [entry.scan_dict for entry in self.files]
-
     def __repr__(self) -> str:
         return f"<Files files={self.files}>"
 
@@ -256,82 +279,148 @@ class Files:
         client: GGClient,
         cache: Cache,
         matches_ignore: Iterable[IgnoredMatch],
-        mode_header: str,
-        command_id: Optional[str] = None,
+        scan_context: ScanContext,
         ignored_detectors: Optional[Set[str]] = None,
-        on_file_chunk_scanned: Callable[
-            [List[Dict[str, Any]]], None
-        ] = lambda chunk: None,
+        progress_callback: Callable[..., None] = lambda advance: None,
+        scan_threads: int = 4,
     ) -> Results:
-        if command_id is None:
-            command_id = generate_command_id()
-        logger.debug("self=%s command_id=%s", self, command_id)
-        cache.purge()
-        scannable_list = self.scannable_list
+        logger.debug("self=%s command_id=%s", self, scan_context.command_id)
+        scanner = Scanner(
+            client, cache, matches_ignore, scan_context, ignored_detectors
+        )
+        return scanner.scan(self.files, progress_callback, scan_threads)
+
+
+class Scanner:
+    def __init__(
+        self,
+        client: GGClient,
+        cache: Cache,
+        matches_ignore: Iterable[IgnoredMatch],
+        scan_context: ScanContext,
+        ignored_detectors: Optional[Set[str]] = None,
+    ):
+        self.client = client
+        self.cache = cache
+        self.matches_ignore = matches_ignore
+        self.ignored_detectors = ignored_detectors
+        self.headers = get_headers(scan_context)
+
+    def _scan_chunk(
+        self, executor: concurrent.futures.ThreadPoolExecutor, chunk: List[File]
+    ) -> Future:
+        """
+        Sends a chunk of files to scan to the API
+        """
+        # `documents` is a version of `chunk` suitable for `GGClient.multi_content_scan()`
+        documents = [{"document": x.document, "filename": x.filename} for x in chunk]
+        return executor.submit(
+            self.client.multi_content_scan,
+            documents,
+            self.headers,
+        )
+
+    def _start_scans(
+        self,
+        executor: concurrent.futures.ThreadPoolExecutor,
+        files: Iterable[File],
+        progress_callback: Callable[..., None],
+    ) -> Dict[Future, List[File]]:
+        """
+        Start all scans, return a tuple containing:
+        - a mapping of future to the list of files it is scanning
+        - a list of files which we did not send to scan because we could not decode them
+
+        on_file
+        """
+        chunks_for_futures = {}
+        skipped_chunk = []
+
+        chunk: List[File] = []
+        for file in files:
+            if file.document:
+                chunk.append(file)
+                if len(chunk) == MULTI_DOCUMENT_LIMIT:
+                    future = self._scan_chunk(executor, chunk)
+                    progress_callback(advance=len(chunk))
+                    chunks_for_futures[future] = chunk
+                    chunk = []
+            else:
+                skipped_chunk.append(file)
+        if chunk:
+            future = self._scan_chunk(executor, chunk)
+            progress_callback(advance=len(chunk))
+            chunks_for_futures[future] = chunk
+        progress_callback(advance=len(skipped_chunk))
+        return chunks_for_futures
+
+    def _collect_results(
+        self,
+        chunks_for_futures: Dict[Future, List[File]],
+    ) -> Results:
+        """
+        Receive scans as they complete, report progress and collect them and return
+        a Results.
+        """
+        self.cache.purge()
+
         results = []
         errors = []
-        chunks = []
-        for i in range(0, len(scannable_list), MULTI_DOCUMENT_LIMIT):
-            chunks.append(scannable_list[i : i + MULTI_DOCUMENT_LIMIT])
+        for future in concurrent.futures.as_completed(chunks_for_futures):
+            chunk = chunks_for_futures[future]
 
-        context = click.get_current_context(silent=True)
+            exception = future.exception()
+            if exception is None:
+                scan = future.result()
+            else:
+                scan = Detail(detail=str(exception))
+                errors.append(
+                    Error(
+                        files=[(file.filename, file.filemode) for file in chunk],
+                        description=scan.detail,
+                    )
+                )
 
-        headers = get_extra_headers(context, command_id=command_id)
-        headers["mode"] = mode_header
+            if not scan.success:
+                handle_scan_chunk_error(scan, chunk)
+                continue
 
-        with concurrent.futures.ThreadPoolExecutor(
-            max_workers=min(CPU_COUNT, 4), thread_name_prefix="content_scan"
-        ) as executor:
-            future_to_scan = {
-                executor.submit(
-                    client.multi_content_scan,
-                    chunk,
-                    headers,
-                ): chunk
-                for chunk in chunks
-            }
-
-            for future in concurrent.futures.as_completed(future_to_scan):
-                chunk = future_to_scan[future]
-                on_file_chunk_scanned(chunk)
-
-                exception = future.exception()
-                if exception is None:
-                    scan = future.result()
-                else:
-                    scan = Detail(detail=str(exception))
-                    errors.append(
-                        Error(
-                            files=[
-                                (file["filename"], file["filemode"]) for file in chunk
-                            ],
-                            description=scan.detail,
+            for file, scanned in zip(chunk, scan.scan_results):
+                remove_ignored_from_result(scanned, self.matches_ignore)
+                remove_results_from_ignore_detectors(scanned, self.ignored_detectors)
+                if scanned.has_policy_breaks:
+                    for policy_break in scanned.policy_breaks:
+                        self.cache.add_found_policy_break(policy_break, file.filename)
+                    results.append(
+                        Result(
+                            content=file.document,
+                            scan=scanned,
+                            filemode=file.filemode,
+                            filename=file.filename,
                         )
                     )
 
-                if not scan.success:
-                    handle_scan_chunk_error(scan, chunk)
-                    continue
-
-                for index, scanned in enumerate(scan.scan_results):
-                    remove_ignored_from_result(scanned, matches_ignore)
-                    remove_results_from_ignore_detectors(scanned, ignored_detectors)
-                    if scanned.has_policy_breaks:
-                        for policy_break in scanned.policy_breaks:
-                            cache.add_found_policy_break(
-                                policy_break, chunk[index]["filename"]
-                            )
-                        results.append(
-                            Result(
-                                content=chunk[index]["document"],
-                                scan=scanned,
-                                filemode=chunk[index]["filemode"],
-                                filename=chunk[index]["filename"],
-                            )
-                        )
-        cache.save()
+        self.cache.save()
         return Results(results=results, errors=errors)
 
+    def scan(
+        self,
+        files: Iterable[File],
+        progress_callback: Callable[..., None],
+        scan_threads: int = 4,
+    ) -> Results:
+
+        with concurrent.futures.ThreadPoolExecutor(
+            max_workers=scan_threads, thread_name_prefix="content_scan"
+        ) as executor:
+            chunks_for_futures = self._start_scans(
+                executor,
+                files,
+                progress_callback,
+            )
+
+            return self._collect_results(chunks_for_futures)
+
 
 class CommitInformation(NamedTuple):
     author: str
@@ -375,7 +464,7 @@ def _parse_patch(
             document = diff[end_of_headers + 1 :]
 
             file_size = len(document.encode("utf-8"))
-            if file_size > MAX_FILE_SIZE * 0.90:
+            if file_size > DOCUMENT_SIZE_THRESHOLD_BYTES * 0.90:
                 continue
 
             if document:
@@ -454,3 +543,46 @@ class Commit(Files):
 
     def __repr__(self) -> str:
         return f"<Commit sha={self.sha} files={self.files}>"
+
+
+def handle_scan_chunk_error(detail: Detail, chunk: List[File]) -> None:
+    logger.error("status_code=%d detail=%s", detail.status_code, detail.detail)
+    if detail.status_code == 401:
+        raise click.UsageError(detail.detail)
+    if detail.status_code is None:
+        raise click.ClickException(
+            f"Error scanning, network error occurred: {detail.detail}"
+        )
+
+    details = None
+
+    display_error("\nError scanning. Results may be incomplete.")
+    try:
+        # try to load as list of dicts to get per file details
+        details = literal_eval(detail.detail)
+    except Exception:
+        pass
+
+    if isinstance(details, list) and details:
+        # if the details had per file details
+        display_error(
+            f"Add the following {pluralize('file', len(details))}"
+            " to your paths-ignore:"
+        )
+        for i, inner_detail in enumerate(details):
+            if inner_detail:
+                click.echo(
+                    f"- {format_text(chunk[i].filename, STYLE['filename'])}:"
+                    f" {str(inner_detail)}",
+                    err=True,
+                )
+        return
+    else:
+        # if the details had a request error
+        filenames = ", ".join([file.filename for file in chunk])
+        display_error(
+            "The following chunk is affected:\n"
+            f"{format_text(filenames, STYLE['filename'])}"
+        )
+
+        display_error(str(detail))
